""" Generator functions to evaluate trial and test functions

NB: Basis evaluation is only needed for the trial function argument in jacobians, as the
multiplication with the test function is part of the sum factorization kernel.
"""

from dune.perftool.generation import (backend,
                                      domain,
                                      get_counter,
                                      get_global_context_value,
                                      iname,
                                      instruction,
                                      kernel_cached,
                                      temporary_variable,
                                      )
from dune.perftool.sumfact.amatrix import (AMatrix,
                                           basis_functions_per_direction,
                                           construct_amatrix_sequence,
                                           name_theta,
                                           quadrature_points_per_direction,
                                           )
from dune.perftool.sumfact.sumfact import (get_facedir,
                                           setup_theta,
                                           SumfactKernel,
                                           sumfact_iname,
                                           sum_factorization_kernel,
                                           )
from dune.perftool.sumfact.quadrature import quadrature_inames
from dune.perftool.sumfact.switch import (get_facedir,
                                          get_facemod,
                                          )
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           )
from dune.perftool.loopy.buffer import initialize_buffer
from dune.perftool.pdelab.driver import FEM_name_mangling
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.tools import maybe_wrap_subscript

from pytools import product

from loopy.match import Writes

import pymbolic.primitives as prim


def name_sumfact_base_buffer():
    count = get_counter('sumfact_base_buffer')
    name = "buffer_{}".format(str(count))
    return name


@kernel_cached
def pymbolic_trialfunction_gradient(element, restriction, component, visitor):
    rawname = "gradu" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)

    # Get a temporary for the gradient
    from ufl.functionview import select_subelement
    sub_element = select_subelement(element, component)
    rank = len(sub_element.value_shape()) + 1
    shape = sub_element.value_shape() + (world_dimension(),)
    shape_impl = ('arr',) * rank
    temporary_variable(name, shape=shape, shape_impl=shape_impl)

    dim = world_dimension()
    buffers = []
    insn_dep = None
    for i in range(dim):
        # Construct the matrix sequence for this sum factorization
        a_matrices = construct_amatrix_sequence(derivative=i,
                                                facedir=get_facedir(restriction),
                                                facemod=get_facemod(restriction),
                                                )

        # Get the vectorization info. If this happens during the dry run, we get dummies
        from dune.perftool.sumfact.vectorization import get_vectorization_info
        a_matrices, buffer, input, index = get_vectorization_info(a_matrices, restriction)

        # Initialize the buffer for the sum fact kernel
        shape = (product(mat.cols for mat in a_matrices),)
        if index is not None:
            shape = shape + (4,)
        input = initialize_buffer(buffer,
                                  base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                                  num=2
                                  ).get_temporary(shape=shape,
                                                  name=input,
                                                  )
        if insn_dep is None:
            insn_dep = frozenset({Writes(input)})

        # Setup the input!
        setup_theta(input, element, restriction, component, index)

        # Add a sum factorization kernel that implements the
        # evaluation of the gradients of basis functions at quadrature
        # points (stage 1)
        var, insn_dep = sum_factorization_kernel(a_matrices,
                                                 buffer,
                                                 1,
                                                 preferred_position=i,
                                                 insn_dep=insn_dep,
                                                 restriction=restriction,
                                                 )
        buffers.append(var)

    # Check whether we want to return early with something that has the indexing
    # already handled! This happens with vectorization when the index coincides
    # with the position in the vector register.
    if index:
        assert len(visitor.indices) == 1
        return maybe_wrap_subscript(var, tuple(prim.Variable(i) for i in quadrature_inames()) + visitor.indices), None

    # TODO this should be quite conditional!!!
    for i, buf in enumerate(buffers):
        # Write solution from sumfactorization to gradient variable
        from pymbolic.primitives import Subscript, Variable
        from dune.perftool.generation import get_backend
        assignee = Subscript(Variable(name), i)
        expression = Subscript(buf, tuple(Variable(i) for i in quadrature_inames()))
        instruction(assignee=assignee,
                    expression=expression,
                    forced_iname_deps=frozenset(get_backend("quad_inames")()),
                    forced_iname_deps_is_final=True,
                    )

    return prim.Variable(name), visitor.indices


@kernel_cached
def pymbolic_trialfunction(element, restriction, component, visitor):
    # Get geometric dimension
    dim = world_dimension()

    # Construct the matrix sequence for this sum factorization
    a_matrices = construct_amatrix_sequence(facedir=get_facedir(restriction),
                                            facemod=get_facemod(restriction),)

    # Get the vectorization info. If this happens during the dry run, we get dummies
    from dune.perftool.sumfact.vectorization import get_vectorization_info
    a_matrices, buffer, input, index = get_vectorization_info(a_matrices, restriction)

    # Flip flop buffers for sumfactorization
    shape = (product(mat.cols for mat in a_matrices),)
    if index is not None:
        shape = shape + (4,)
    initialize_buffer(buffer,
                      base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                      num=2
                      ).get_temporary(shape=shape,
                                      name=input,
                                      )

    # Setup the input!
    setup_theta(input, element, restriction, component, index)

    # Add a sum factorization kernel that implements the evaluation of
    # the basis functions at quadrature points (stage 1)
    var, _ = sum_factorization_kernel(a_matrices,
                                      buffer,
                                      1,
                                      preferred_position=None,
                                      insn_dep=frozenset({Writes(input)}),
                                      outshape=tuple(mat.rows for mat in a_matrices if mat.rows != 1),
                                      restriction=restriction,
                                      )

    if index:
        index = (index,)
    else:
        index = ()

    return prim.Subscript(var,
                          tuple(prim.Variable(i) for i in quadrature_inames()) + index
                          ), visitor.indices


@iname
def sumfact_lfs_iname(bound, dim):
    name = "sumfac_lfs_{}".format(dim)
    domain(name, bound)
    return name


@backend(interface="lfs_inames", name="sumfact")
def lfs_inames(element, restriction, number=1, context=''):
    assert number == 1
    dim = world_dimension()
    return tuple(sumfact_lfs_iname(basis_functions_per_direction(), d) for d in range(dim))


@backend(interface="evaluate_basis")
@kernel_cached
def evaluate_basis(element, name, restriction):
    temporary_variable(name, shape=())
    theta = name_theta()
    quad_inames = quadrature_inames()
    inames = lfs_inames(element, restriction)
    facedir = get_facedir(restriction)

    # Collect the pairs of lfs/quad inames that are in use
    # On facets, the normal direction of the facet is excluded
    prod = tuple(prim.Subscript(prim.Variable(theta),
                                (prim.Variable(i), prim.Variable(j))
                                )
                 for (i, j) in zip(quad_inames, tuple(iname for i, iname in enumerate(inames) if i != facedir))
                 )

    # Add the missing direction on facedirs by evaluating at either 0 or 1
    if facedir is not None:
        facemod = get_facemod(restriction)
        from dune.perftool.sumfact.amatrix import PolynomialLookup, name_polynomials
        prod = prod + (prim.Call(PolynomialLookup(name_polynomials(), False),
                                 (prim.Variable(inames[facedir]), facemod)),)

    # Issue the product
    instruction(expression=prim.Product(prod),
                assignee=prim.Variable(name),
                forced_iname_deps=frozenset(quad_inames + inames),
                forced_iname_deps_is_final=True,
                )


def pymbolic_basis(element, restriction, number):
    assert number == 1
    assert element.num_sub_elements() == 0

    name = "phi_{}".format(FEM_name_mangling(element))
    name = restricted_name(name, restriction)
    evaluate_basis(element, name, restriction)

    return prim.Variable(name)


@backend(interface="evaluate_grad")
@kernel_cached
def evaluate_reference_gradient(element, name, restriction):
    dim = world_dimension()
    temporary_variable(name, shape=(dim,))
    quad_inames = quadrature_inames()
    inames = lfs_inames(element, restriction)
    facedir = get_facedir(restriction)

    # Map the direction to a quadrature iname
    quadinamemapping = {}
    i = 0
    for d in range(local_dimension()):
        if d == facedir:
            i = i + 1
        quadinamemapping[i] = quad_inames[d]
        i = i + 1

    for d in range(dim):
        prod = []
        for i in range(dim):
            if i != facedir:
                prod.append(prim.Subscript(prim.Variable(name_theta(derivative=d == i)),
                                           (prim.Variable(quadinamemapping[i]), prim.Variable(inames[i]))
                                           ))
        if facedir is not None:
            facemod = get_facemod(restriction)
            from dune.perftool.sumfact.amatrix import PolynomialLookup, name_polynomials
            prod.append(prim.Call(PolynomialLookup(name_polynomials(), facedir == d),
                                  (prim.Variable(inames[facedir]), facemod)),)

        assignee = prim.Subscript(prim.Variable(name), (d,))
        expression = prim.Product(tuple(prod))

        instruction(assignee=assignee,
                    expression=expression,
                    forced_iname_deps=frozenset(quad_inames + inames),
                    forced_iname_deps_is_final=True,
                    )


def pymbolic_reference_gradient(element, restriction, number):
    assert number == 1

    # TODO: Change name?
    name = "js_{}".format(FEM_name_mangling(element))
    name = restricted_name(name, restriction)
    evaluate_reference_gradient(element, name, restriction)

    return prim.Variable(name)
